%%% Algorithm: The master process defines and creates:
%%% - 'CWorker' pcs worker processes which execute 'Fun' (the given fun expression)
%%% - a receiver process which collects the result of the workers; 
%%% then runs them in parallel. 
%%%
%%% The receiver process waits for all the workers, then 
%%% sends back the accumulated results to the master process.
%%%
%%% To build: Execute c(pp1). command in an Erlang shell
%%% To run: Execute pp1:run_pp(fun()->do_some_task end, 10). command in the same shell
%%%
%%% Example:
%%% 1>c(pp1).
%%% 2>pp1:run_pp(fun()->io:format("Worker process (~p)~n",[self()]), do_some_task end,10).
%%% Worker process (<0.63.0>)
%%% Worker process (<0.64.0>)
%%% Worker process (<0.65.0>)
%%% Worker process (<0.66.0>)
%%% Worker process (<0.67.0>)
%%% Worker process (<0.68.0>)
%%% Worker process (<0.69.0>)
%%% Worker process (<0.70.0>)
%%% Worker process (<0.71.0>)
%%% Worker process (<0.72.0>)
%%% [do_some_task,do_some_task,do_some_task,do_some_task,
%%%  do_some_task,do_some_task,do_some_task,do_some_task,
%%%  do_some_task,do_some_task]


-module(pp1).
-author("Viktória Fördős").
-export([run_pp/2]).

% Usage: pp1:run_pp(fun()->do_some_task end, 10).
run_pp(Fun, CWorker) when is_integer(CWorker) andalso CWorker>0 
    andalso is_function(Fun)->
    % the unique identifier of the receiver process is returned
    ReceiverPid = produce_receiver(self(), CWorker),
    % Worker process definition
    WorkerFun=fun()->
                      % execute 'Fun'
                      Result = Fun(),
                      % send back the result as a message to the receiver
                      ReceiverPid ! {worker_result, Result}                   
              end,
    % Worker process is started CWorker times
    [spawn(F) || F <-lists:duplicate(CWorker, WorkerFun) ],
    % The accumulated result as a message is sent back,
    % the master process waits for it and receives it.
    receive
        {receiver_finished, R} ->R
    end.

produce_receiver(ParentPid, CWorker) when is_pid(ParentPid) ->
    % Receiver process is started as a separate process,
    % and it's identifier is returned
    spawn(fun()->
                  % Receiver process definition
                  % The process waits for all the worker
                  Result = receive_loop(CWorker, []),
                  % then sends back the accumulated results 
                  % to the master process.
                  ParentPid ! {receiver_finished, Result}
          end).

receive_loop(0, Acc)->
    Acc;
receive_loop(CWorker, AccIn) when is_integer(CWorker) andalso CWorker>0 ->
    % the process is being blocked until one of the worker has finished 
    % and sends back its result
    Result = receive
        {worker_result, R} -> R
    end,
    % The process has to wait for the other workers to finish (CWorker-1 times)
    receive_loop(CWorker-1, AccIn ++ [Result]).